@@ -20,9 +20,6 @@
 import sys
 import os
 import stat
-import time
-import string
-import math
 import threading
 import Queue
 
@@ -34,306 +31,22 @@ import locale
 import gettext
 
 import puppy
+from FileSystemModel import *
+from util import *
 
 APP_NAME = 'Guppy'
-VERSION = '0.0.2'
+VERSION = '0.0.3'
 AUTHORS = ['Tony Tsui tsui.tony@gmail.com']
 WEBSITE = 'http://guppy.nongnu.org'
 COPYRIGHT = 'Copyright (C) 2005 Tony Tsui'
 LICENSE = 'GNU Public License'
 
-
-def humanReadableSize(size):
-	div_count = 0
-	new_size = size
-	prev_size = new_size
-	
-	while new_size > 0 and new_size > 1000:
-		div_count += 1
-		prev_size = new_size
-		new_size = new_size/1024
-	
-	if prev_size > 1000:
-		# Divide by float for more precise number
-		human_size = "%.1f" % (prev_size / 1024.0)
-	else:
-		human_size = str(prev_size)
-	
-	if div_count == 1:
-		human_size += ' KB'
-	elif div_count == 2:
-		human_size += ' MB'
-	elif div_count == 3:
-		human_size += ' GB'
-	else:
-		human_size += ' B'
-
-	return human_size
-
-def convertToBytes(size):
-	size, unit = size.split()
-
-	size = float(size)
-	
-	if unit == 'GB':
-		size = size * math.pow(1024, 3)
-	elif unit == 'MB':
-		size = size * math.pow(1024, 2)
-	elif unit == 'KB':
-		size = size * 1024
-		
-	return size
-
-class FileSystemModel(gtk.ListStore):
-	TYPE_COL, ICON_COL, NAME_COL, DATE_COL, SIZE_COL = range(5)
-	
-
-	def __init__(self):
-		self.current_dir = None
-		gtk.ListStore.__init__(self, gobject.TYPE_STRING, gobject.TYPE_STRING,
-		                             gobject.TYPE_STRING, gobject.TYPE_STRING,
-		                             gobject.TYPE_STRING)
-
-	def find(self, name):
-		for row in self:
-			if name == row[FileSystemModel.NAME_COL]:
-				return row
-				
-		return None
-			
-	def getCWD(self):
-		return self.current_dir
-	
-	def sort_func(self, model, iter1, iter2, col=None):
-		type1 = model.get_value(iter1, FileSystemModel.TYPE_COL)
-		type2 = model.get_value(iter2, FileSystemModel.TYPE_COL)
-		
-		if type1 == type2:
-			if col == FileSystemModel.NAME_COL:
-				return self.string_sort_func(model, iter1, iter2, col)
-			elif col == FileSystemModel.DATE_COL:
-				return self.date_sort_func(model, iter1, iter2, col)
-			elif col == FileSystemModel.SIZE_COL:
-				return self.size_sort_func(model, iter1, iter2, col)
-			else:
-				# FIXME: Raise exception here
-				print "ERROR: Unknown column type: ", col
-				return 0
-		elif type1 == 'f' and type2 == 'd':
-			return 1
-		else:
-			return -1
-	
-	def string_sort_func(self, model, iter1, iter2, col=None):
-		string1 = string.lower(model.get_value(iter1, col))
-		string2 = string.lower(model.get_value(iter2, col))
-
-		if string1 == string2:
-			return 0;
-		elif string1 < string2:
-			return -1
-		else:
-			return 1
-
-	def date_sort_func(self, model, iter1, iter2, col=None):
-		# Date column is empty for '..' directory
-		date1_str = model.get_value(iter1, col)
-		if len(date1_str) == 0:
-			return -1
-		
-		date2_str = model.get_value(iter2, col)
-		if len(date2_str) == 0:
-			return 1
-
-		format = '%a %b %d %Y'
-		date1 = time.strptime(date1_str, format)
-		date2 = time.strptime(date2_str, format)
-
-		if date1 == date2:
-			return self.string_sort_func(model, iter1, iter2, FileSystemModel.NAME_COL);
-		elif date1 < date2:
-			return -1
-		else:
-			return 1
-
-	def size_sort_func(self, model, iter1, iter2, col=None):
-		size1 = model.get_value(iter1, col).split()
-		size2 = model.get_value(iter2, col).split()
-
-		size1_len = len(size1)
-		size2_len = len(size2)
-		
-		if size1_len < size2_len:
-			return -1
-		elif size1_len > size2_len:
-			return 1
-		else:
-			return 0
-
-		unit_weight = { 'B' : 1, 'KB' : 2, 'MB' : 3, 'GB' : 4 }
-
-		# Sizes in bytes may not have 'B' unit postfix
-		if len(size1) < 2:
-			size1[1] = unit_weight['B']
-		else:
-			size1[1] = unit_weight[size1[1]]
-
-		return 1
-		# Sizes in bytes may not have 'B' unit appended
-		if len(size2) < 2:
-			size2[1] = unit_weight['B']
-		else:
-			size2[1] = unit_weight[size2[1]]
-			
-		if size1[1] == size2[1]:
-			size1[0] = float(size1[0])
-			size2[0] = float(size2[0])
-			if size1[0] == size2[0]:
-				return self.string_sort_func(model, iter1, iter2, FileSystemModel.NAME_COL);
-			elif size1[0] < size2[0]:
-				return -1
-			else:
-				return 1
-		elif size1[1] < size2[1]:
-			return -1
-		else:
-			return 1
-
-
-class PVRFileSystemModel(FileSystemModel):
-	dir_sep = '\\'
-	def __init__(self):
-		FileSystemModel.__init__(self)
-
-		# FIXME: Get dir from when Guppy last exited
-		self.current_dir = ''
-
-		self.freespace = 0
-				
-		self.puppy = puppy.Puppy()
-		
-		self.changeDir()
-
-
-	def changeDir(self, dir=None):
-		if dir:
-			# PVR root dir path is an empty string. Using '\' character does not
-			# work.
-			if dir == '\\':
-				self.current_dir = ''
-			elif len(self.current_dir) > 0 and self.current_dir[-1] != '\\':
-				dir = self.current_dir + '\\' + dir
-		else:
-			dir = self.current_dir
-
-		norm_path = os.path.normpath(dir.replace('\\', '/'))
-		if norm_path != '.':
-			dir = norm_path.replace('/', '\\')
-		else:
-			dir = ''
-
-		# We can't get a listing during a file transfer
-		try:
-			pvr_files = self.puppy.listDir(dir)
-		except puppy.PuppyBusyError:
-			msg = _('Can not change directory during a file transfer')
-			dialog = gtk.MessageDialog(type=gtk.MESSAGE_INFO,
-			                           buttons=gtk.BUTTONS_CLOSE,
-			                           message_format=msg)
-			response = dialog.run()
-			dialog.destroy()
-
-			return
-
-		self.current_dir = dir
-		
-		# Clear model
-		if len(self) > 0:
-			self.clear()
-
-		for file in pvr_files:
-			# TODO: Set icon based on file type. Use dummy icon for now
-			if file[FileSystemModel.TYPE_COL] == 'd':
-				file.insert(FileSystemModel.ICON_COL, gtk.STOCK_DIRECTORY)
-			else:				
-				file.insert(FileSystemModel.ICON_COL, gtk.STOCK_FILE)
-				
-			file[FileSystemModel.SIZE_COL] = humanReadableSize(int(file[FileSystemModel.SIZE_COL]))
-			self.append(file)
-	
-	def freeSpace(self):
-		try:
-			total, free = self.puppy.getDiskSpace()
-			self.freespace = free
-		except:
-			pass
-
-		return humanReadableSize(self.freespace)
-
-class PCFileSystemModel(FileSystemModel):
-	def __init__(self):
-		FileSystemModel.__init__(self)
-		
-		# FIXME: Get dir from when Guppy last exited
-		self.current_dir = os.environ['HOME']
-		
-		self.changeDir()
-		
-	def changeDir(self, dir=None):
-		if dir:
-			if dir[0] != '/':
-				dir = self.current_dir + '/' + dir
-		else:
-			dir = self.current_dir
-
-		dir = os.path.normpath(dir)
-		
-		if not os.access(dir, os.F_OK):
-			return
-
-		self.current_dir = dir
-			
-		if len(self) > 0:
-			self.clear()
-		
-		# Parent directory
-		self.append(['d', gtk.STOCK_DIRECTORY, '..', '', ''])
-		
-		for file in os.listdir(self.current_dir):
-			mode = os.stat(self.current_dir + '/' + file)
-
-			if stat.S_ISDIR(mode[stat.ST_MODE]):
-				type = 'd'
-				icon = gtk.STOCK_DIRECTORY
-				size = ''
-			else:
-				type = 'f'
-				icon = gtk.STOCK_FILE
-				size = humanReadableSize(mode[stat.ST_SIZE])
-			
-			mtime = time.strftime('%a %b %d %Y', time.localtime(mode[stat.ST_MTIME]))
-			entry = [ type, icon, file, mtime, size ]
-			self.append(entry)
-
-	def freeSpace(self):
-		cmd = 'df ' + self.current_dir
-		pipe = os.popen(cmd)
-
-		# Skip Headers
-		pipe.readline()
-		output = pipe.readline().split()
-
-		exit_stat = pipe.close()
-		if exit_stat != None and os.WEXITSTATUS(exit_stat) != 0:
-			# TODO: Raise exception
-			print "ERROR: Failed to get disk free space (`%s')" % (cmd)
-			return None
-		
-		# Multiple by 1024 to convert from kilobytes to bytes
-		return humanReadableSize(int(output[3])*1024)
 			
 class GuppyWindow:
-	FREE_SPACE_UPDATE_INTERVAL = 5000 # 5 secs
+	SCREEN_INFO_UPDATE_INTERVAL = 10 * 60 * 1000 # 10 mins (unit in milliseconds)
+	
+	# Quit command to put on transfer queue
+	QUIT_CMD = 'Quit'
 	
 	def __init__(self):	
 		# Find out proper way to find glade files
@@ -366,8 +79,11 @@ class GuppyWindow:
 		self.pvr_model = PVRFileSystemModel()
 		self.pc_model = PCFileSystemModel()
 		
-		self.free_space_timeout_id = gobject.timeout_add(GuppyWindow.FREE_SPACE_UPDATE_INTERVAL, self.update_free_space)
-
+		self.updateFreeSpace()
+		
+		# Timer to update screen info
+		gobject.timeout_add(GuppyWindow.SCREEN_INFO_UPDATE_INTERVAL, self.update_screen_info)
+		
 		# Queue to put files to be transferred. The transfer thread gets files
 		# to transfer from this queue.
 		self.transfer_queue = Queue.Queue(0)
@@ -375,7 +91,7 @@ class GuppyWindow:
 		# Queue to put all files which has been transferred. Files on this queue
 		# are removed when the Transfer Frame Clear button is clicked.
 		self.transfer_complete_queue = Queue.Queue(0)
-		
+
 		# Create thread to transfer files
 		self.transfer_thread = TransferThread(self)
 		self.transfer_thread.setDaemon(True)
@@ -540,19 +256,42 @@ class GuppyWindow:
 	def on_download_btn_clicked(self, widget, data=None):
 		self.transferFile('download')
 
-	def on_guppy_window_delete_event(self, widget, event, data=None):
+	def on_window_delete_event(self, widget, event, data=None):
 		self.on_quit(widget, data)
+		# Return True to stop event from propagating to default handler
+		# which destroys the window.
+		return True
 		
 	def on_path_entry_activate(self, widget, fs_model):
 		fs_model.changeDir(widget.get_text())
 		
-	def on_quit(self, widget, data=None):
-		gtk.main_quit()
-		self.puppy.cancelTransfer()
+		if fs_model == self.pc_model:
+			self.updateFreeSpace(self.pc_model)
 		
+	def on_quit(self, widget, data=None):
+		# Empty out transfer queue so the TransferThread can't get another
+		# FileTransfer object after we cancel the current transfer.
+		while True:
+			try:
+				file_transfer = self.transfer_queue.get_nowait()
+			except Queue.Empty:
+				break
+				
+			file_transfer.cancel()
+			
+		# Stop transfer if one is in progress
+		self.puppy.cancelTransfer()
+
+		# Add QUIT command to transfer queue.
+		self.transfer_queue.put(GuppyWindow.QUIT_CMD)
+
+		# We don't call gtk.main_quit() here because the TransferThread may try
+		# to call some gtk functions. Instead if the TransferThread gets the
+		# QUIT_CMD from the transfer queue it will call gtk.main_quit().
+				
 	def on_show_file_transfer_toggled(self, widget, data=None):
 		transfer_frame = self.glade_xml.get_widget('transfer_frame')
-		if not transfer_frame.get_property('visible'):
+		if widget.get_active():
 			self.show_transfer_frame()
 		else:
 			transfer_frame.hide()
@@ -562,10 +301,22 @@ class GuppyWindow:
 		self.pc_liststore.get_model().refilter()
 		self.pvr_liststore.get_model().refilter()
 
+	def on_transfer_clear_btn_clicked(self, widget, data=None):
+		# Loop until we get a Queue.Empty exception
+		while True:
+			try:
+				file_transfer = self.transfer_complete_queue.get_nowait()
+			except Queue.Empty:
+				break
+
+			progress_box = file_transfer.xml.get_widget('progress_box')
+			# progress_box may be None if the Remove button was clicked on it
+			if progress_box:
+				progress_box.destroy()
+	
 	def on_treeview_changed(self, widget, fs_model):
 		model, files = widget.get_selected_rows()
 		
-		file_no = 1
 		file_count = len(files)
 		total_size = 0
 		for path in files:
@@ -603,33 +354,17 @@ class GuppyWindow:
 		if type == 'd':
 			fs_model.changeDir(name)
 			path = fs_model.getCWD()
-			if isinstance(fs_model, PCFileSystemModel):
+			if fs_model == self.pc_model:
 				self.pc_path_entry.set_text(path)
+				self.updateFreeSpace(self.pc_model)
 			else:
 				self.pvr_path_entry.set_text(path)
 			
 	def on_turbo_toggled(self, widget, data=None):
 		self.puppy.setTurbo(widget.get_active())
 		
-	def on_upload_btn_clicked(self, widget, data=None):
-		self.transferFile('upload')
-
-	def on_transfer_clear_btn_clicked(self, widget, data=None):
-		# Loop until we get a Queue.Empty exception
-		while True:
-			try:
-				file_transfer = self.transfer_complete_queue.get_nowait()
-			except Queue.Empty:
-				break
-
-			progress_box = file_transfer.xml.get_widget('progress_box')
-			# progress_box may be None if the Remove button was clicked on it
-			if progress_box:
-				progress_box.destroy()
-	
 	def on_transfer_close_btn_clicked(self, widget, data=None):
 		self.show_file_transfer_action.set_active(False)
-		self.on_show_file_transfer_toggled(widget)
 
 	def on_transfer_remove_btn_clicked(self, widget, file_transfer):
 		file_transfer.cancel()
@@ -640,13 +375,18 @@ class GuppyWindow:
 	def on_transfer_stop_btn_clicked(self, widget, data=None):
 		self.puppy.cancelTransfer()
 
+	def on_upload_btn_clicked(self, widget, data=None):
+		self.transferFile('upload')
+
+	def reallyQuit(self):
+		gtk.main_quit()
 	
 	def show_transfer_frame(self):
 		# Set position of pane separator
 		# Give progress frame height of 180 pixels. This should be enough room
 		# to show two file transfers.
 		vpane = self.glade_xml.get_widget('vpane')
-		vpane.set_position(vpane.get_allocation().height - 185)
+		vpane.set_position(vpane.get_allocation().height - 188)
 		
 		transfer_frame = self.glade_xml.get_widget('transfer_frame')
 		transfer_frame.show()
@@ -685,10 +425,11 @@ class GuppyWindow:
 				return
 
 		queue_box = self.glade_xml.get_widget('queue_vbox')
-		self.show_transfer_frame()
+		
+		# Show File Transfer pane
+		self.show_file_transfer_action.set_active(True)
 
-		file_count = len(files)
-		file_no = 1
+		existing_files = []
 		for path in files:
 			iter = model.get_iter(path)
 			file = model.get_value(iter, FileSystemModel.NAME_COL)
@@ -698,26 +439,13 @@ class GuppyWindow:
 				dst_dir = self.pc_model.getCWD()
 				src_file = src_dir + '\\' + file
 				dst_file = dst_dir + '/' + file
-				existing_file = os.access(dst_file, os.F_OK)
+				file_exists = os.access(dst_file, os.F_OK)
 			else:
 				src_dir = self.pc_model.getCWD()
 				dst_dir = self.pvr_model.getCWD()
 				src_file = src_dir + '/' + file
 				dst_file = dst_dir + '\\' + file
-				existing_file = self.pvr_model.find(file)
-
-			if existing_file:
-				msg = _('The file') + ' "' + dst_file + '" ' + _('already exists. Would you like to replace it?')
-				msg2 = _('If you replace an existing file, its contents will be overwritten.')
-				dialog = gtk.MessageDialog(type=gtk.MESSAGE_QUESTION,
-				                           buttons=gtk.BUTTONS_YES_NO,
-				                           message_format=msg)
-				response = dialog.run()
-				dialog.destroy()
-
-				if response == gtk.RESPONSE_NO or response == gtk.RESPONSE_DELETE_EVENT:
-					file_no += 1
-					continue
+				file_exists = self.pvr_model.find(file)
 
 			xml = gtk.glade.XML('guppy.glade', 'progress_box')
 			xml.signal_autoconnect(self)		
@@ -734,8 +462,6 @@ class GuppyWindow:
 			file_label.set_text(file)
 			from_label.set_text(src_dir)
 			to_label.set_text(dst_dir)
-
-			queue_box.pack_start(progress_box, expand=False)
 	
 			file_transfer = FileTransfer(direction, src_file, dst_file, xml)
 
@@ -743,26 +469,79 @@ class GuppyWindow:
 			remove_btn = xml.get_widget('transfer_remove_button')
 			remove_btn.connect('clicked', self.on_transfer_remove_btn_clicked, file_transfer)
 
-			self.transfer_queue.put(file_transfer, True, None)
+			if file_exists:
+				existing_files.append(file_transfer)
+			else:
+				queue_box.pack_start(progress_box, expand=False)
+				self.transfer_queue.put(file_transfer, True, None)
 
+		# Confirm with user that they wish to overwrite existing files
+		replace_all = False
+		for file in existing_files:
+			if replace_all == False:
+				SKIP, REPLACE, REPLACE_ALL = range(3)
+				msg = _('The file') + ' "' + file.dst + '" ' + _('already exists. Would you like to replace it?')
+				msg2 = _('If you replace an existing file, its contents will be overwritten.')
+				dialog = gtk.MessageDialog(type=gtk.MESSAGE_QUESTION,
+				                           message_format=msg)
+				                           
+				if len(existing_files) > 1:
+					dialog.add_button(_('Replace All'), REPLACE_ALL)
+				dialog.add_buttons( _('Skip'), SKIP, _('Replace'), REPLACE)
+				response = dialog.run()
+				dialog.destroy()
+	
+				if response == SKIP or response == gtk.RESPONSE_DELETE_EVENT:
+					continue
+					
+				if response == REPLACE_ALL:
+					replace_all = True
 
-	def update_treeviews(self):		
+			progress_box = file.xml.get_widget('progress_box')
+			queue_box.pack_start(progress_box, expand=False)
+			self.transfer_queue.put(file, True, None)
+
+	def updateFreeSpace(self, fs_model=None):			
+		'''Update label showing free space available on each file system.
+	
+		fs_model  Model to update free space. Default is None which updates both
+		          file system.
+		'''
+		if fs_model == self.pvr_model or fs_model == None:
+			self.pvr_free_space_label.set_text(_('Free Space') + ': ' + self.pvr_model.freeSpace())
+			
+		if fs_model == self.pc_model or fs_model == None:
+			self.pc_free_space_label.set_text(_('Free Space') + ': ' + self.pc_model.freeSpace())
+
+	def update_screen_info(self):
+		# Update amount of free space available
+		self.updateFreeSpace()
+		
+		# Update treeviews.
+		self.updateTreeViews()
+		return True
+
+	def updateTreeViews(self):		
 		# Update FileSystemModel view				
 		models = [ (self.pc_model, self.pc_treeview), (self.pvr_model, self.pvr_treeview) ]
 		for model, treeview in models:
 			handler_id = treeview.get_data('changed_handler_id')
 			selection = treeview.get_selection()
+			
+			# Store currently selected rows because they will be wiped when
+			# the data model is updated.
 			selected_rows = selection.get_selected_rows()
+			
 			selection.handler_block(handler_id)
+			# Update PVR file system cache
+			if isinstance(model, PVRFileSystemModel):
+				model.updateCache()
 			model.changeDir()
 			selection.handler_unblock(handler_id)
+			
+			# Reselect rows
 			for path in selected_rows[1]:
 				selection.select_path(path)
-			
-	def update_free_space(self):
-		self.pvr_free_space_label.set_text(_('Free Space') + ': ' + self.pvr_model.freeSpace())
-		self.pc_free_space_label.set_text(_('Free Space') + ': ' + self.pc_model.freeSpace())
-		return True
 
 
 		
@@ -796,10 +575,19 @@ class TransferThread(threading.Thread):
 
 	def run(self):
 		while True:
+				
 			file_transfer = self.file_queue.get(True, None)
+
+			# Check if guppy is quiting
+			if file_transfer == GuppyWindow.QUIT_CMD:
+				gtk.gdk.threads_enter()
+				self.guppy.reallyQuit()
+				gtk.gdk.threads_leave()
+				return
+				
 			if not file_transfer.isAlive():
 				continue
-				
+
 			direction = file_transfer.direction
 			src_file = file_transfer.src
 			dst_file = file_transfer.dst
@@ -825,8 +613,6 @@ class TransferThread(threading.Thread):
 			# Set width of Stop button to be the same as Remove button
 			gtk.gdk.threads_enter()
 			allocation = remove_btn.get_allocation()
-			gtk.gdk.threads_leave()
-			gtk.gdk.threads_enter()
 			stop_btn.set_size_request(allocation.width, allocation.height)
 			gtk.gdk.threads_leave()
 
@@ -871,32 +657,24 @@ class TransferThread(threading.Thread):
 					
 				gtk.gdk.threads_enter()
 				progress_bar.set_fraction(float(percent)/100)
-				gtk.gdk.threads_leave()
-				
-				gtk.gdk.threads_enter()
 				progress_bar.set_text('(' + time['remaining'] + ' ' + _('Remaining') + ')')
-				gtk.gdk.threads_leave()
-				
-				gtk.gdk.threads_enter()
+
 				while gtk.events_pending():
 					gtk.main_iteration()
 				gtk.gdk.threads_leave()
 			
+			gtk.gdk.threads_enter()
 			if transfer_successful:
-				gtk.gdk.threads_enter()
 				progress_bar.set_fraction(1)
-				gtk.gdk.threads_leave()
-				
-				gtk.gdk.threads_enter()
+
 				progress_bar.set_text(_('Finished'))
-				gtk.gdk.threads_leave()
 			else:
-				gtk.gdk.threads_enter()
 				progress_bar.set_text(_('Transfer Failed'))
-				gtk.gdk.threads_leave()
+
+			gtk.gdk.threads_leave()
 				
 
-			# Insensitise all widgets			
+			# Desensitise all widgets			
 			for widget in ['progress_hbox1', 'progress_hbox2']:
 				widget = xml.get_widget(widget)
 				gtk.gdk.threads_enter()
@@ -916,7 +694,7 @@ class TransferThread(threading.Thread):
 			
 			# Update file treeviews
 			gtk.gdk.threads_enter()
-			self.guppy.update_treeviews()
+			self.guppy.updateTreeViews()
 			gtk.gdk.threads_leave()
 			
 			# Put on queue for completed transfers
